!to "neoload.prg",cbm
!convtab pet

* = $0801
!byte $0c,$08,$0a,$00,$9e,$32,$30,$36,$31,$00,$00,$00

synclen     = 5     ; Length in bytes of synchronization sequence
conflen     = 8     ; Length in bytes of confirmation sequence
blockwords  = $2d   ; Number of words per block

bittab      = $0f00 ; bittab+x contains number of bits in byte x
blockbuf    = $1000 ; Address of block buffer (4 * blockwords bytes)
blocknr     = $10b4 ; 2-byte number of the next block expected

dataword    = $a8   ; 3-byte data word for hamming codes
parity      = $ab   ; 1-byte parity word for hamming codes
corrected   = $b4   ; 3-byte corrected data word
bitcount    = $57   ; (overlap) 1-byte count for number of matching bits
checksum    = $57   ; (overlap) 4-byte checksum
bitbuf      = $fb   ; 5-byte bit buffer

!zone start

start
    sei                 ; Don't move this instruction!

    jsr createtab       ; Create bit count table

    lda #$10            ; Wait for cassette button press
    bit $01
    bne *-2

    lda #$00
    sta $d020           ; Set border color to black
    sta blocknr+0
    sta blocknr+1       ; The first block we expect is nr 0

    jsr initload

    ; "target" is set to $000700 by default, change if necessary.
    ; Here we set the georam page initially, it will be updated when necessary
    ; during loading.
    lda target+1
    sta $dffe
    lda target+2
    sta $dfff

.loadblock
    jsr sync
    bcc .end

    jsr readblock
    jsr decode

    lda blockbuf+$80
    cmp blocknr+0
    bne .fail
    lda blockbuf+$81
    cmp blocknr+1
    bne .fail           ; Ensure that no block was missed

    inc blocknr+0
    bne *+5
    inc blocknr+1

    jsr check           ; Compute checksum

    ldx #3
.check
    lda checksum,x
    cmp blockbuf+$83,x
    bne .fail
    dex
    bpl .check          ; Compare checksum to the one computed

    ldy #0
    ldx target+0
.copy
    lda blockbuf,y
.offset
    sta $de00,x
    inx
    bne .noinc
    inc target+1
    bit target+1
    bvc .incdone
    inc target+2
.incdone
    lda target+1
    and #$3f
    sta target+1
    sta $dffe
    lda target+2
    sta $dfff
.noinc
    iny
    bpl .copy
    stx target+0

    jmp .loadblock

.fail
    jsr finishload
    jmp 64738

.end
    jsr finishload
    lda #0
    sta $dffe
    sta $dfff
    jmp 57000

target
    !byte $00, $07, $00

; Compute checksum of data block in buffer

!zone check
check
    ldx #0
    ldy #0
    stx checksum+0
    stx checksum+1
    stx checksum+2
    stx checksum+3
.checkblock
    lda checksum,x
    eor blockbuf,y
    sta checksum,x
    inx
    iny
    cpx #4
    bne .checkblock
    jsr mix
    ldx #0
    cpy #$80
    bne .checkblock

    ldx #2
.checkadr
    lda checksum,x
    eor blockbuf+$80,x
    sta checksum,x
    dex
    bpl .checkadr
    jmp mix

; Decode the 135 bytes of data in blockbuf

!zone decode
decode
    ldx #0
    ldy #0
.word
    stx .savex+1
    sty .savey+1

    lda blockbuf+0,x
    sta corrected+0
    eor #$ff
    sta dataword+0
    lda blockbuf+1,x
    sta corrected+1
    eor #$ff
    sta dataword+1
    lda blockbuf+2,x
    sta corrected+2
    eor #$ff
    sta dataword+2      ; Load data bytes

    jsr calc_parity     ; Compute parity bits
    ldy .savey+1
    eor blockbuf+blockwords*3,y
    and #$1f
    beq .ok             ; Skip correction if no errors are detected
    tax
    jsr correct_word
.ok
.savex
    ldx #0
.savey
    ldy #0

    lda corrected+0
    sta blockbuf,x
    inx
    lda corrected+1
    sta blockbuf,x
    inx
    lda corrected+2
    sta blockbuf,x
    inx

    iny
    cpy #blockwords
    bne .word
    rts

initload
    ldy #0              ; Disable VIC-II interrupts
    sty $d01a
    lda $d019           ; Acknowledge VIC-II interrupts
    sta $d019

    lda #$7f            ; Disable CIA interrupts
    sta $dc0d
    sta $dd0d
    lda $dc0d           ; Acknowledge CIA interrupts
    lda $dd0d

    lda #$1d            ; Enable motor
    sta $01

    lda #$0b            ; Disable screen
    sta $d011

    sty $dd05           ; Y = 0
    lda #$fe
    sta $dd04           ; Timer A: $00fe

    rts

finishload
    lda #$1b            ; Enable screen
    sta $d011

    lda #$37            ; Disable motor
    sta $01

    lda #$81
    sta $dc0d           ; Enable CIA interrupts

    lda #$0e            ; Restore border color
    sta $d020

    cli
    rts

; Read interleaved bit data into 45 29-bit words, where the first 45 3-byte
; blocks (135 bytes in total) contain the low 24 bits of each word, and the
; low 5 bits of the following 45 bytes contain the high bits of each word.
; NOTE: the high 3 bits of the last 45 bytes contain garbage.

!zone readblock
readblock
    ldy #$18            ; 24 low bits per word
    ldx #0              ; Start at offset 0
.readlow
    jsr readbit
    rol blockbuf,x      ; Write bit to buffer location
    inx
    inx
    inx                 ; Advance to same bit of next word
    cpx #3*blockwords   ; Past the buffer area?
    bcc .readlow
    txa
    sbc #3*blockwords+1
    tax
    bpl *+4
    ldx #2              ; X := (X - 3*blockwords - 1) mod 3
    dey
    bne .readlow        ; Loop until all low bits in the words have been read

    ldy #5              ; 5 high bits per word
    ldx #0              ; Start at offset 0
.readhigh
    jsr readbit
    rol blockbuf+3*blockwords,x
    inx                 ; Advance to next word
    cpx #blockwords     ; Past the buffer area?
    bne .readhigh
    ldx #0              ; Go back to start of buffer
    dey
    bne .readhigh       ; Loop until all high bits in the words have been read

    rts

; Read X bits from tape into A. If X is more than 8, only the last 8 bits will
; be returned. On return, A contains the word (most significant bit first) and
; X is 0. Y untouched.

!zone readword
readbyte
    ldx #8
readword
    lda #0
.continue
    pha
    jsr readbit
    pla
    rol
    dex
    bne .continue
    rts

; Wait for a synchronization sequence.
; If the carry flag is set on return, synchronization is complete and data
; will follow. Otherwise this indicates the end of the block stream.

!zone sync
sync
    lda #$ff
    sta bitbuf+0
    sta bitbuf+1
    sta bitbuf+2
    sta bitbuf+3        ; Initialize buffer, to avoid junk data from before

.sync
    jsr readbit
    rol bitbuf+0
    rol bitbuf+1
    rol bitbuf+2
    rol bitbuf+3        ; bitbuf = (bitbuf << 1) + bit

    lda #0
    sta bitcount
    clc

    ; You have about 185 cycles to spare for the rolling above + the loop,
    ; which together account for 30*n cycles per pattern byte. This does not
    ; account for time used by readbit, so 5 is probably the largest
    ; reasonable value for synclen
    ldx #synclen-1
.syncbyte
    lda .pattern,x      ; 4
    eor bitbuf,x        ; 8
    tay                 ; 10
    lda bitcount        ; 13
    adc bittab,y        ; 17
    sta bitcount        ; 20
    dex                 ; 22
    bpl .syncbyte       ; 25

    cmp #10
    bcs .sync           ; If too many bits are wrong, continue looking

    lda #0
    sta bitcount
    clc

    ldy #conflen-1
.confirm
    jsr readbyte
    eor .magic,y
    tax
    lda bittab,x
    adc bitcount
    sta bitcount
    dey
    bpl .confirm

    cmp #12             ; Fewer than 12 bits wrong?
    bcc .done           ; Exit with carry = 0
    cmp #(conflen*8)-11 ; Ferew than 12 bits right?
    bcc sync            ; Exit with carry = 1

.done
    rts

.magic                  ; Magic bytes for confirmation
    !byte $24, $93, $4a, $c8, $e0, $f5, $d5, $de
.pattern                ; Pattern expected before start of data
    !byte $8e, $95, $8d, $a6, $00


!zone createtab
createtab
    ldy #0              ; Y = index in table
.byte
    tya
    ldx #0              ; X = number of 1 bits
.loop
    asl                 ; Get one bit
    beq .done           ; If there are no more 1 bits, abort
    bcc .loop           ; X += bit
    inx
    bne .loop           ; This should always branch
.done
    bcc *+3             ; X += bit (in case we had $80 before asl)
    inx

    txa
    sta bittab,y        ; Store bit count

    iny
    bne .byte           ; Loop for $00 to $ff
    rts

; Read one bit from tape, return it in the carry flag.
; A is destroyed, X and Y untouched

!zone readbit
readbit
    lda #$10
.wait
    bit $dc0d
    beq .wait           ; Wait until a bit has been read
    lda $dd0d           ; Check if timer has finished
    lsr                 ; Bit value to carry flag
    lda #$19
    sta $dd0e           ; Restart timer
    inc $d020           ; Switch border color
    rts

; Correct a bit in the 3-byte dataword. The received parity bits XOR the
; correct ones go in X. Only call this if the XOR is non-zero. If a parity bit
; was corrupted no action is taken.

!zone correct_word
correct_word
    dex
    lda .table,x        ; Check which bit the error occured in
    bmi .done           ; Negative value means a parity bit, so just return
    tay
    and #$07            ; A = corrupt bit number within byte
    tax
    tya
    lsr
    lsr
    lsr
    tay                 ; Y = byte with corrupt bit
    lda corrected,y
    eor .shift,x        ; Invert bit in dataword, and we're done
    sta corrected,y
.done
    rts
.shift
    !byte $01, $02, $04, $08, $10, $20, $40, $80
.table
    !byte $ff, $ff, $00, $ff, $01, $02, $03, $ff
    !byte $04, $05, $06, $07, $08, $09, $0a, $ff
    !byte $0b, $0c, $0d, $0e, $0f, $10, $11, $12
    !byte $13, $14, $15, $16, $17

; Calculate 5 Hamming code parity bits from the 3-byte word in dataword.
; Note that the dataword should be inverted (i.e. all bytes EOR #$FF).
; Upon return, the parity bits are stored in the low 5 bits of parity (and A).
; Note that dataword is destroyed.

!zone calc_parity
calc_parity
    ldx #0
    stx parity
.loop
    lda #0              ; 2
    lsr dataword+0      ; 7
    sbc #0              ; 9     A <- $00 if CF = 1, $ff if CF = 0
    and .table,x        ; 13    Only XOR with relevant bits
    eor parity          ; 16
    sta parity          ; 19

    lda #0              ; 21
    lsr dataword+1      ; 26
    sbc #0              ; 28
    and .table+8,x      ; 32
    eor parity          ; 35
    sta parity          ; 38

    lda #0              ; 40
    lsr dataword+2      ; 45
    sbc #0              ; 47
    and .table+16,x     ; 51
    eor parity          ; 54
    sta parity          ; 57

    inx                 ; 59
    cpx #8              ; 61
    bne .loop           ; 64
    rts

.table
    !byte $03, $05, $06, $07, $09, $0a, $0b, $0c
    !byte $0d, $0e, $0f, $11, $12, $13, $14, $15
    !byte $16, $17, $18, $19, $1a, $1b, $1c, $1d

; Mix the 4 bytes in "checksum"
mix:
    clc
    lda checksum+0
    eor #$e2

    eor checksum+2
    ror
    ror
    ror
    adc checksum+2
    sta checksum+2

    eor checksum+1
    rol
    rol
    rol
    adc checksum+1
    sta checksum+1

    eor checksum+3
    ror
    ror
    ror
    ror
    adc checksum+3
    sta checksum+3

    eor checksum+0
    rol
    rol
    adc checksum+0
    sta checksum+0

    eor checksum+3
    rol
    rol
    adc checksum+3
    sta checksum+3

    eor checksum+1
    ror
    ror
    ror
    adc checksum+1
    sta checksum+1

    eor checksum+0
    ror
    ror
    ror
    ror
    adc checksum+0
    sta checksum+0

    eor checksum+2
    adc checksum+2
    sta checksum+2

    rts

