#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define BUFSZ	1048576	/* buffer size = 1 MB */
#define FREQ	44100
#define WAVHDSZ	44	/*!< Size of .WAV file header. */

struct tap_t
{
	unsigned char *tmem;	/* storage for the loaded tap. */
	int len;		
};

unsigned char checksum=0;
unsigned long int time=0;
/**********************************************************************************/
static int s_out(unsigned char amp, int finished, FILE *filep, unsigned long *bp, char *outbuf)
{
	if (!finished) {
		outbuf[(*bp)++] = amp;
		if (*bp == BUFSZ) {				/* buffer full? */
			fwrite(outbuf, 1, BUFSZ, filep);	/* add full buffer to the file */
			*bp = 0;
		}
		return 1;

	/* push all remaining vaautaurlid bytes to the file if 'finished' is true.. */

	} else {

		fwrite(outbuf, 1, *bp, filep);
		return 0;
	}
}

static int drawwavesquare(int len, int amp, char signd, FILE *filep, unsigned long *bp, char *outbuf)
{
	int x, y, off;
	int samples = 0;

	if (signd)
		off = 0;
	else 

		off = 128;

	for (x = 0; x < len; x++) {
		if (x < (len >> 1))	/* peak first */
			y = amp;
		else
			y =- amp;
			samples += s_out(y + off, 0, filep, bp, outbuf);
	}

	return samples;
}

/*
 * write out sine wave data of amplitude 'amp' and length 'len'.
 * if parameter 'signd' is set (1) the wave data will be in a signed format
 * else unsigned.
 */
/**********************************************************************************/
static int drawwavesine(int len, int amp, char signd, FILE *filep, unsigned long *bp, char *outbuf)
{
	int x, y, off;
	float deg, inc;
	float rads = 180 / 3.141592654;
	int samples = 0;

	if (signd)
		off = 0;
	else
		off = 128;

	for (x = 0; x < len; x++) {
		inc = (float)360 / len;
		deg = (x * inc);	/* +180 makes the wave peak first */
		y = amp * sin(deg / rads);
		samples += s_out(y + off, 0, filep, bp, outbuf);
	}

	return samples;
}
/**********************************************************************************/
static int
wav_write(unsigned char *tap, int taplen, const char *filename, char sine)
{
	long i, lv, ccnt;
	int b, tapversion;
	FILE *fp;
	unsigned long bpos = 0;
	char *outbuf;
	double fratio = FREQ / 985248.0;
	unsigned int dsize, totalsamples = 0;
	char wavhd[WAVHDSZ] = { 82, 73, 70, 70, 0, 0, 0, 0, 87, 65, 86, 69,
				102, 109, 116, 32, 16, 0, 0, 0, 1, 0, 1, 0,
				68, 172, 0, 0, 68, 172, 0, 0, 1, 0, 8, 0,
				100, 97, 116, 97, 0, 0, 0, 0 };

	tapversion = tap[12];
 
	fp = fopen(filename, "w+b");	/* create output file... */
	if (fp == NULL)
		return - 1;

	outbuf = (char*)malloc(BUFSZ);	/* create output buffer... */
	if (outbuf == NULL)
		return - 1;
 
	/* write out a dummy header, proper one needs to go in after. */

	for (i = 0; i < WAVHDSZ; i++)
		fputc(0, fp);
 
	/* write out the waveform data... */

	for(i = 20; i < taplen; i++) {
		b = tap[i];
 
		/* write TAP v0 pause... */

		if (b == 0 && tapversion == 0) {
			lv = floor(20000 * fratio);
			if (sine)
				totalsamples += drawwavesine(lv, 0, 0, fp, &bpos, outbuf);
			else
				totalsamples += drawwavesquare(lv, 0, 0, fp, &bpos, outbuf);
		}
 
	/* write TAP v1 pause... */

	if (b == 0 && tapversion == 1) {
		ccnt = (tap[i + 1]) + (tap[i + 2] << 8) + (tap[i + 3] << 16);
		i += 3;
		lv = floor(ccnt * fratio);
		if (sine)
			totalsamples += drawwavesine(lv, 0, 0, fp, &bpos, outbuf);
		else
			totalsamples += drawwavesquare(lv, 0, 0, fp, &bpos, outbuf);
	}
 
	/* write out non-pause... */

	if (b != 0) {
		lv = floor((b * 8) * fratio);
		if (sine)
			totalsamples += drawwavesine(lv, 127, 0, fp, &bpos, outbuf);
		else
			totalsamples += drawwavesquare(lv, 127, 0, fp, &bpos, outbuf);
		}
	}

	totalsamples += s_out(0, 1, fp, &bpos, outbuf);	/* add all valid remaining buffer bytes. */
 
	/* write FileLen-8 to wav header... */

	dsize = totalsamples + WAVHDSZ - 8;
	for (i = 0; i < 4; i++) {
		b = (dsize & (0xFF << (i * 8))) >> (i * 8);
		wavhd[4 + i] = b;
	}

	/* write DataChunkLen to wav header... */
   
	dsize = totalsamples;
	for (i = 0; i < 4; i++) {
		b = ((dsize & (0xFF << (i * 8)))) >> (i * 8);
		wavhd[40 + i] = b;
	}
 
	/* write header array to file and clean up... */

	rewind(fp);
	for (i = 0; i < WAVHDSZ; i++)
		fputc(wavhd[i], fp);
 
	fclose(fp);
	free(outbuf);
 
	return 0;
}


static int tap2wav(char *tapfilename,char *wavfilename)
{
	FILE *fp;
	int flen;
	struct tap_t tap;

	fp = fopen(tapfilename, "rb");
	if (fp == NULL) return -1;

	fseek(fp, 0, SEEK_END);
	flen = ftell(fp);
	rewind(fp);

	/* Allocate enough space to load the file into */
	tap.tmem = (unsigned char*)malloc(flen);
	if(tap.tmem == NULL) {
		printf("\nError: malloc failed in load_tap().");
		fclose(fp);
		return 0;
	}

	/* Load data into buffer */
	fread(tap.tmem, flen, 1, fp);
	fclose(fp);

	/* Set the 'tap' structure file length subfield */
	tap.len = flen;	
// 	printf("%d\n",tap.len);
	wav_write(tap.tmem, tap.len, wavfilename,0);
	return 0;
}
/**********************************************************************************/
int main(int argc, char **argv) {
	if(argc != 3) {
		fprintf(stderr, "Usage: %s <infile.tap> <outfile.wav>\n",
			argv[0]);
		return 1;
	}

	tap2wav(argv[1], argv[2]);
	return 0;
}


